/*
 * fb-contrib - Auxiliary detectors for Java programs
 * Copyright (C) 2005-2009 Dave Brosius
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.mebigfatguy.fbcontrib.detect;

import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.generic.Type;

import edu.umd.cs.findbugs.BugInstance;
import edu.umd.cs.findbugs.BugReporter;
import edu.umd.cs.findbugs.BytecodeScanningDetector;

/**
 * looks for methods that call a method to retrieve a reference to an object, 
 * to use to load a constant. It is simpler and more performant to access the
 * static variable directly from the class itself.
 */
public class NeedlessInstanceRetrieval extends BytecodeScanningDetector
{
	private static final int SEEN_NOTHING = 0;
	private static final int SEEN_INVOKE = 1;
	private static final int SEEN_POP = 2;
	
	private BugReporter bugReporter;
    private LineNumberTable lnTable;
	private int state;
    private int invokePC;
	/**
     * constructs a NIR detector given the reporter to report bugs on
     * @param bugReporter the sync of bug reports
	 */
	public NeedlessInstanceRetrieval(BugReporter bugReporter) {
		this.bugReporter = bugReporter;
	}
	
    /**
     * overrides the interface to collect the line number table, and reset state
     * 
     * @param obj the content object of the currently parsed code
     */
    @Override
    public void visitCode(Code obj) {
        try {
            lnTable = obj.getLineNumberTable();
            if (lnTable != null) {
                state = SEEN_NOTHING;
                super.visitCode(obj);
            }
        } finally {
            lnTable = null;
        }
    }
    
	/**
	 * overrides the interface to find accesses of static variables off of an instance
	 * immediately fetched from a method call.
	 * 
	 * @param seen the opcode of the currently visited instruction
	 */
	@Override
	public void sawOpcode(int seen) {
		switch (state)
		{
			case SEEN_NOTHING:
				if ((seen == INVOKEINTERFACE)
				||  (seen == INVOKEVIRTUAL)) {
					String sig = getSigConstantOperand();
					Type retType = Type.getReturnType(sig);
					if (retType.getSignature().startsWith("L")) {
						String clsName = getClassConstantOperand();
						if (!"java/lang/Object".equals(clsName)
						&&  !"java/lang/String".equals(clsName)) {
	                        invokePC = getPC();
							state = SEEN_INVOKE;
						}
                    }
				}
			break;
			
			case SEEN_INVOKE:
				if (seen == POP)
					state = SEEN_POP;
				else
					state = SEEN_NOTHING;
			break;
			
			case SEEN_POP:
				if ((seen >= ACONST_NULL) && (seen <= DCONST_1)) {
                    if (lnTable.getSourceLine(invokePC) == lnTable.getSourceLine(getPC())) {
    					bugReporter.reportBug(new BugInstance(this, "NIR_NEEDLESS_INSTANCE_RETRIEVAL", NORMAL_PRIORITY)
    									.addClass(this)
    									.addMethod(this)
    									.addSourceLine(this));
                    }
				}
				state = SEEN_NOTHING;
			break;
            
            default:
                state = SEEN_NOTHING;
            break;
		}
	}
}
 