/*
 * fb-contrib - Auxiliary detectors for Java programs
 * Copyright (C) 2005-2009 Dave Brosius
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.mebigfatguy.fbcontrib.detect;

import java.util.HashSet;
import java.util.Set;

import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.Type;

import com.mebigfatguy.fbcontrib.utils.Integer14;

import edu.umd.cs.findbugs.BugInstance;
import edu.umd.cs.findbugs.BugReporter;
import edu.umd.cs.findbugs.BytecodeScanningDetector;
import edu.umd.cs.findbugs.ba.ClassContext;

/**
 * Looks for methods that store the return result in a local variable, and
 * then immediately returns that local variable.
 */
public class UnnecessaryStoreBeforeReturn extends BytecodeScanningDetector
{
	private static final int SEEN_NOTHING = 0;
	private static final int SEEN_STORE = 1;
	private static final int SEEN_LOAD = 2;
	private static final Set<Integer> branchInstructions = new HashSet<Integer>();
	static {
		branchInstructions.add(Integer14.valueOf(GOTO));
		branchInstructions.add(Integer14.valueOf(GOTO_W));
		branchInstructions.add(Integer14.valueOf(IFEQ));
		branchInstructions.add(Integer14.valueOf(IFNE));
		branchInstructions.add(Integer14.valueOf(IFLT));
		branchInstructions.add(Integer14.valueOf(IFGE));
		branchInstructions.add(Integer14.valueOf(IFGT));
		branchInstructions.add(Integer14.valueOf(IFLE));
		branchInstructions.add(Integer14.valueOf(IF_ICMPEQ));
		branchInstructions.add(Integer14.valueOf(IF_ICMPNE));
		branchInstructions.add(Integer14.valueOf(IF_ICMPLT));
		branchInstructions.add(Integer14.valueOf(IF_ICMPGE));
		branchInstructions.add(Integer14.valueOf(IF_ICMPGT));
		branchInstructions.add(Integer14.valueOf(IF_ICMPLE));
		branchInstructions.add(Integer14.valueOf(IF_ACMPEQ));
		branchInstructions.add(Integer14.valueOf(IF_ACMPNE));
		branchInstructions.add(Integer14.valueOf(IFNULL));
		branchInstructions.add(Integer14.valueOf(IFNONNULL));
	}
	
	private final BugReporter bugReporter;
	private Set<Integer> branchTargets;
	private Set<Integer> catchTargets;
	private int state;
	private int storeReg;
	
	/**
     * constructs a USBR detector given the reporter to report bugs on
     * @param bugReporter the sync of bug reports
	 */
	public UnnecessaryStoreBeforeReturn(BugReporter bugReporter) {
		this.bugReporter = bugReporter;
	}
	
	/**
	 * implements the visitor to create and clear the branchTargets
	 * 
	 * @param classContext the context object for the currently parsed class
	 */
	@Override
	public void visitClassContext(ClassContext classContext) {
		try {
			branchTargets = new HashSet<Integer>();
			catchTargets = new HashSet<Integer>();
			super.visitClassContext(classContext);
		} finally {
			branchTargets = null;
			catchTargets = null;
		}
	}
	/**
	 * implements the visitor to make sure method returns a value, and then clears the targets
	 * 
	 * @param obj the context object of the currently parsed code block
	 */
	@Override
	public void visitCode(Code obj) {
		Method m = getMethod();
		String sig = m.getSignature();
		Type t = Type.getReturnType(sig);
		if (!t.equals(Type.VOID)) {
			state = SEEN_NOTHING;
			branchTargets.clear();
			CodeException[] ces = obj.getExceptionTable();
			catchTargets.clear();
			for (CodeException ce : ces) {
				if (ce.getCatchType() != 0) {
					catchTargets.add(Integer14.valueOf(ce.getHandlerPC()));
				}
			}
			super.visitCode(obj);
		}
	}
	
	/**
	 * implements the visitor to look for store of registers immediately before returns
	 * of that register
	 * 
	 * @param seen the opcode of the currently parsed instruction
	 */
	@Override
	public void sawOpcode(int seen) {
        switch (state) {
			case SEEN_NOTHING:
				if (!catchTargets.contains(Integer14.valueOf(getPC()))) {
					if (lookForStore(seen))
						state = SEEN_STORE;
				}
			break;
			
			case SEEN_STORE:
				if (branchTargets.contains(Integer14.valueOf(getPC()))) {
					state = SEEN_NOTHING;
					break;
				}

				state = lookForLoad(seen) ? SEEN_LOAD : SEEN_NOTHING;
			break;
			
			case SEEN_LOAD:
				if ((seen >= IRETURN) && (seen <= ARETURN)) {
					bugReporter.reportBug(new BugInstance(this, "USBR_UNNECESSARY_STORE_BEFORE_RETURN", NORMAL_PRIORITY)
						.addClass(this)
						.addMethod(this)
						.addSourceLine(this));
				}
				state = SEEN_NOTHING;
			break;
		}
		
		if (branchInstructions.contains(Integer14.valueOf(seen))) {
			branchTargets.add(Integer14.valueOf(getBranchTarget()));
		}
		
	}
	
	/**
	 * checks if the current opcode is a store, if so saves the register
	 * 
	 * @param seen the opcode of the currently parsed instruction
	 * @return if a store was seen
	 */
	private boolean lookForStore(int seen) {
		if ((seen >= ISTORE) && (seen <= ASTORE))
			storeReg = getRegisterOperand();
		else if ((seen >= ISTORE_0) && (seen <= ISTORE_3))
			storeReg = seen - ISTORE_0;
		else if ((seen >= LSTORE_0) && (seen <= LSTORE_3))
			storeReg = seen - LSTORE_0;
		else if ((seen >= FSTORE_0) && (seen <= FSTORE_3))
			storeReg = seen - FSTORE_0;
		else if ((seen >= DSTORE_0) && (seen <= DSTORE_3))
			storeReg = seen - DSTORE_0;
		else if ((seen >= ASTORE_0) && (seen <= ASTORE_3))
			storeReg = seen - ASTORE_0;
		else
			return false;
		return true;
	}
	
	/**
	 * looks for a load of the register that was just stored
	 * 
	 * @param seen the opcode of the currently parsed instruction
	 * @return if the load was seen
	 */
	private boolean lookForLoad(int seen) {
		int loadReg;
		if ((seen >= ILOAD) && (seen <= ALOAD))
			loadReg = getRegisterOperand();
		else if ((seen >= ILOAD_0) && (seen <= ILOAD_3))
			loadReg = seen - ILOAD_0;
		else if ((seen >= LLOAD_0) && (seen <= LLOAD_3))
			loadReg = seen - LLOAD_0;
		else if ((seen >= FLOAD_0) && (seen <= FLOAD_3))
			loadReg = seen - FLOAD_0;
		else if ((seen >= DLOAD_0) && (seen <= DLOAD_3))
			loadReg = seen - DLOAD_0;
		else if ((seen >= ALOAD_0) && (seen <= ALOAD_3))
			loadReg = seen - ALOAD_0;
		else
			return false;
		
		return (storeReg == loadReg);
	}
}
